"""A simple sql builder based on standard Python type hints"""

__version__ = '0.4.2'
__all__ = [
    "BaseSqlify",
    "Sqlite3Sqlify",
    "Psycopg2Sqlify",
    "Session",
    "Fetch",
    "Order",
    "DatabaseType",
    "Fetch",
    "Order",
    "DatabaseType",
]

from .builder import BaseSqlify, Sqlite3Sqlify, Psycopg2Sqlify
from .operators import SqlOperator, RawSQL, DecreaseSQL, IncreaseSQL
from .session import Session
from .value_objects import Fetch, Order, DatabaseType



from unittest import TestCase, mock

from sqlify import Psycopg2Sqlify, RawSQL, IncreaseSQL, DecreaseSQL, Order


class TestBuilder(TestCase):
    table_name = "test_table"

    def setUp(self):
        self.cursor = mock.MagicMock()
        self.logger = mock.MagicMock()

        self.sqlify = Psycopg2Sqlify(
            self.cursor,
            self.logger,
        )

    def assertQuery(self, sql: str):
        query = str(self.cursor.execute.call_args[0][0]).lower()
        sql = sql.format(
            table=self.table_name,
        )

        self.cursor.execute.assert_called_once()
        self.assertEqual(query, sql)

    def test_fetchall_multiple_field(self):
        self.sqlify.fetchall(self.table_name, fields=["bonus", "ts"])

        self.assertQuery("select bonus, ts from {table}")

    def test_fetchall_single_field(self):
        self.sqlify.fetchall(self.table_name, fields="bonus")

        self.assertQuery("select bonus from {table}")

    def test_fetchone_single_field(self):
        self.sqlify.fetchone(self.table_name, fields="bonus")

        self.assertQuery("select bonus from {table} limit 1")

    def test_fetchall_where_single_condition(self):
        self.sqlify.fetchall(
            self.table_name,
            fields="bonus",
            where=(
                "bonus_key = %(bonus_key)s",
                dict(),
            ),
        )

        self.assertQuery("select bonus from {table} where bonus_key = %(bonus_key)s")

    def test_fetchall_where_list_of_conditions(self):
        self.sqlify.fetchall(
            self.table_name,
            fields="bonus",
            where=(
                [
                    "bonus_key = %(bonus_key)s",
                    "ts = %(ts)s",
                ],
                dict(),
            ),
        )

        self.assertQuery(
            "select bonus from {table} where bonus_key = %(bonus_key)s "
            "and ts = %(ts)s"
        )

    def test_fetchall_where_conditions_limit_and_offset(self):
        self.sqlify.fetchall(
            self.table_name,
            fields="bonus",
            where=(
                [
                    "bonus_key = %(bonus_key)s",
                    "ts = %(ts)s",
                ],
                dict(),
            ),
            limit=5,
            offset=10,
        )

        self.assertQuery(
            "select bonus from {table} where bonus_key = %(bonus_key)s "
            "and ts = %(ts)s limit 5 offset 10"
        )

    def test_fetchall_where_conditions_order_by(self):
        self.sqlify.fetchall(
            self.table_name,
            fields="bonus",
            where=(
                [
                    "bonus_key = %(bonus_key)s",
                    "ts = %(ts)s",
                ],
                dict(),
            ),
            order=("id", Order.ASC),
        )

        self.assertQuery(
            "select bonus from {table} where bonus_key = %(bonus_key)s "
            "and ts = %(ts)s order by id asc"
        )

    def test_fetchall_where_conditions_order_by_desc(self):
        self.sqlify.fetchall(
            self.table_name,
            fields="bonus",
            where=(
                [
                    "bonus_key = %(bonus_key)s",
                    "ts = %(ts)s",
                ],
                dict(),
            ),
            order=("id", Order.DESC),
        )

        self.assertQuery(
            "select bonus from {table} where bonus_key = %(bonus_key)s "
            "and ts = %(ts)s order by id desc"
        )

    def test_fetchall_where_conditions_order_by_desc_raw(self):
        self.sqlify.fetchall(
            self.table_name,
            fields="bonus",
            where=(
                [
                    "bonus_key = %(bonus_key)s",
                    "ts = %(ts)s",
                ],
                dict(),
            ),
            order=("id", "DESC"),
        )

        self.assertQuery(
            "select bonus from {table} where bonus_key = %(bonus_key)s "
            "and ts = %(ts)s order by id desc"
        )

    def test_insert(self):
        self.sqlify.insert(self.table_name, data=dict(asd="test", ts="123"))

        self.assertQuery("insert into {table} (asd, ts) values(%s, %s)")

    def test_insert_returning_everything(self):
        self.sqlify.insert(
            self.table_name,
            data=dict(asd="test", ts="123"),
            returning="*",
        )

        self.assertQuery("insert into {table} (asd, ts) values(%s, %s) returning *")

    def test_update_no_conditions(self):
        self.sqlify.update(self.table_name, data=dict(asd="test", ts="123"))

        self.assertQuery(
            "update test_table set asd = %(asd_datainput)s, ts = %(ts_datainput)s"
        )

    def test_update_with_conditions(self):
        self.sqlify.update(
            self.table_name,
            data=dict(asd="test", ts="123"),
            where=(
                [
                    "bonus_key = %(bonus_key)s",
                    "ts = %(ts)s",
                ],
                dict(),
            ),
        )

        self.assertQuery(
            "update test_table set asd = %(asd_datainput)s, "
            "ts = %(ts_datainput)s "
            "where bonus_key = %(bonus_key)s and ts = %(ts)s"
        )

    def test_update_with_raw_sql_conditions(self):
        self.sqlify.update(
            self.table_name,
            where=(
                ["id = %(id)s", "end_date > now() at time zone 'utc'"],
                dict(id=1),
            ),
            data=dict(
                end_date=RawSQL("now() at time zone 'utc'"),
            ),
        )

        self.assertQuery(
            "update test_table set end_date = now() at time zone 'utc' "
            "where id = %(id)s and end_date > now() at time zone 'utc'"
        )

    def test_update_with_increase_sql(self):
        self.sqlify.update(
            self.table_name,
            data=dict(
                earned=IncreaseSQL(10),
            ),
        )

        self.assertQuery("update test_table set earned = earned + %(earned_datainput)s")

    def test_update_with_decrease_sql(self):
        self.sqlify.update(
            self.table_name,
            data=dict(
                earned=DecreaseSQL(10),
            ),
        )

        self.assertQuery("update test_table set earned = earned - %(earned_datainput)s")

    def test_update_with_conditions_returning_everything(self):
        self.sqlify.update(
            self.table_name,
            data=dict(asd="test", ts="123"),
            where=(
                [
                    "bonus_key = %(bonus_key)s",
                    "ts = %(ts)s",
                ],
                dict(),
            ),
            returning="*",
        )

        self.assertQuery(
            "update test_table set asd = %(asd_datainput)s, ts = %(ts_datainput)s "
            "where bonus_key = %(bonus_key)s and ts = %(ts)s "
            "returning *"
        )

    def test_delete_no_conditions(self):
        self.sqlify.delete(
            self.table_name,
        )

        self.assertQuery("delete from test_table")

    def test_delete_with_conditions(self):
        self.sqlify.delete(
            self.table_name,
            where=(
                [
                    "bonus_key = %(bonus_key)s",
                    "ts = %(ts)s",
                ],
                dict(),
            ),
        )

        self.assertQuery(
            "delete from test_table where bonus_key = %(bonus_key)s and ts = %(ts)s"
        )

    def test_delete_with_conditions_returning_everything(self):
        self.sqlify.delete(
            self.table_name,
            where=(
                [
                    "bonus_key = %(bonus_key)s",
                    "ts = %(ts)s",
                ],
                dict(),
            ),
            returning="*",
        )

        self.assertQuery(
            "delete from test_table where bonus_key = %(bonus_key)s "
            "and ts = %(ts)s returning *"
        )



# -*- coding: utf-8 -*-
from datetime import datetime
from io import StringIO
from logging import Logger
from typing import Optional, List, Tuple, Union, Dict, IO, Any

from .operators import RawSQL, IncreaseSQL, DecreaseSQL, SqlOperator
from .value_objects import Order, Fetch


class BaseSqlify(object):
    connection = None
    logger = None

    def __init__(self, cursor, logger: Logger = None):
        self._cursor = cursor
        self._logger = logger

    @property
    def _unnamed_parameter(self):
        raise NotImplementedError("Database parameter not defined")

    def _format_parameter(self, parameter: str) -> str:
        raise NotImplementedError("Database parameter not defined")

    def _format_where(self, where: Optional[Union[Tuple[Union[List, str], Union[List, Dict]], Union[List, str]]]) \
            -> Optional[Tuple[str, Union[List, Dict]]]:
        if where and len(where) > 0 and isinstance(where[0], list):
            # Where is a list, we must convert it into a string
            where = (" and ".join(where[0]), where[1])

        return where

    def fetchone(
            self,
            table: str,
            fields: Optional[Union[str, List[str]]] = "*",
            where: Optional[Union[str, List[str], Tuple[Union[List[str], str], Union[List, Dict]]]] = None,
            group: Optional[Union[List[str], str]] = None,
            having: Optional[str] = None,
            order: Optional[Union[str, Tuple[str, Union[Order, str]]]] = None,
            offset: int = None,
            with_sq: Optional[Dict[str, str]] = None,
    ) -> Optional[Union[Dict, List]]:
        """Get a single result
        table = (str) table_name
        fields = (field1, field2 ...) list of fields to select
        where = ("parameterized_statement", [parameters])
                eg: ("id=%s and name=%s", [1, "test"])
        order = [field, ASC|DESC]
        """
        conditions, parameters = self._split_where(where)

        sql = self._select(
            table=table,
            fields=fields,
            where=conditions,
            group=group,
            having=having,
            order=order,
            limit=1,
            offset=offset,
            with_sq=with_sq,
        )
        cur = self.execute(sql, parameters)
        return cur.fetchone()

    def fetchall(
            self,
            table: str,
            fields: Optional[Union[str, List[str]]] = "*",
            where: Optional[Union[str, List[str], Tuple[Union[List[str], str], Union[List, Dict]]]] = None,
            group: Optional[Union[List[str], str]] = None,
            having: Optional[str] = None,
            order: Optional[Union[str, Tuple[str, Union[Order, str]]]] = None,
            limit: Optional[int] = None,
            offset: Optional[int] = None,
            with_sq: Optional[Dict[str, str]] = None,
    ) -> Optional[List[Union[Dict, List]]]:
        """Get all results
        table = (str) table_name
        fields = (field1, field2 ...) list of fields to select
        where = ("parameterized_statement", [parameters])
                eg: ("id=%s and name=%s", [1, "test"])
        order = [field, ASC|DESC]
        limit = [limit, offset]
        """
        conditions, parameters = self._split_where(where)

        sql = self._select(
            table=table,
            fields=fields,
            where=conditions,
            group=group,
            having=having,
            order=order,
            limit=limit,
            offset=offset,
            with_sq=with_sq,
        )
        cur = self.execute(sql, parameters)
        return cur.fetchall()

    def insert(
            self,
            table: str,
            data: Dict[str, Union[str, bool, int, datetime]],
            returning: str = None,
    ) -> Optional[Union[Dict, int]]:
        """Insert a record"""
        cols, vals = self._format_insert(data)
        sql = "INSERT INTO {} ({}) VALUES({})".format(table, cols, vals)
        sql += self._returning(returning)
        cur = self.execute(sql, list(data.values()))
        return cur.fetchone() if returning else cur.rowcount

    def update(
            self,
            table: str,
            data: Dict[str, Union[str, bool, int, datetime, SqlOperator]],
            where: Optional[Union[str, List[str], Tuple[Union[List[str], str], Union[List, Dict]]]] = None,
            returning: str = None,
    ) -> Optional[Union[Dict, int]]:
        """Insert a record"""
        conditions, parameters = self._split_where(where)
        query = self._format_update(data)
        arguments = {}
        for key, value in data.items():
            arguments[key + "_datainput"] = value  # TODO

        sql = "UPDATE {} SET {}".format(table, query)
        sql += self._where(conditions) + self._returning(returning)

        if parameters is not None:
            arguments.update(parameters)

        cur = self.execute(sql, arguments)
        return cur.fetchall() if returning else cur.rowcount

    def delete(
            self,
            table: str,
            where: Optional[Tuple[Union[List, str], Union[List, Dict]]] = None,
            returning: str = None,
    ) -> Optional[Union[Dict, int]]:
        """Delete rows based on a where condition"""
        conditions, parameters = self._split_where(where)

        sql = f"DELETE FROM {table}"
        sql += self._where(conditions) + self._returning(returning)
        cur = self.execute(sql, parameters)
        return cur.fetchall() if returning else cur.rowcount

    def execute(
            self,
            sql,
            params=None,
            fetch: Optional[Fetch] = None,
    ) -> Any:
        """Executes a raw query"""
        # self._cursor.timestamp = time.time()
        self._cursor.execute(sql, params or ())
        # self._logger.debug("query", self._cursor.query)

        return self._cursor

    def copy_expert(
            self,
            file: Union[IO, StringIO],
            table: str,
            fields: Optional[Union[str, List[str]]] = "*",
            where: Optional[Tuple[Union[List, str], Union[List, Dict]]] = None,
            order: Optional[Tuple[str, Order]] = None,
            limit: Optional[int] = None,
            offset: Optional[int] = None,
    ) -> IO:
        """Export query to file
        table = (str) table_name
        fields = (field1, field2 ...) list of fields to select
        where = ("parameterized_statement", [parameters])
                eg: ("id=%s and name=%s", [1, "test"])
        order = [field, ASC|DESC]
        limit = [limit, offset]
        """
        conditions, parameters = self._split_where(where)

        sql = self._select(table, fields, conditions, order, limit, offset)
        sql = f"copy ({sql}) to stdout with csv delimiter ',' header"
        sql = self._cursor.mogrify(sql, parameters)

        return self._cursor.copy_expert(sql=sql, file=file)

    def truncate(self, table: str, restart_identity: bool = False, cascade: bool = False) -> None:
        """Truncate a table or set of tables
        db.truncate('tbl1')
        db.truncate('tbl1, tbl2')
        """
        sql = f"TRUNCATE {table}"
        if restart_identity:
            sql += " RESTART IDENTITY"
        if cascade:
            sql += " CASCADE"
        self.execute(sql)

    def drop(self, table: str, cascade: bool = False) -> None:
        """Drop a table"""
        sql = f"DROP TABLE IF EXISTS {table}"
        if cascade:
            sql += " CASCADE"
        self.execute(sql)

    def create(self, table: str, schema: str) -> None:
        """Create a table with the schema provided
        pg_db.create('my_table','id SERIAL PRIMARY KEY, name TEXT')"""
        self.execute("CREATE TABLE {} ({})".format(table, schema))

    def commit(self) -> None:
        """Commit a transaction"""
        self._cursor.connection.commit()

    def rollback(self) -> None:
        """Roll-back a transaction"""
        self._cursor.connection.rollback()

    def _format_insert(self, data):
        """Format insert dict values into strings"""
        cols = ", ".join(data.keys())
        vals = ", ".join([self._unnamed_parameter for k in data])

        return cols, vals

    def _format_update(self, data):
        """Format update dict values into string"""
        arguments = []
        for key, value in data.items():
            if isinstance(value, RawSQL):
                arguments.append(f"{key} = {value}")
            elif isinstance(value, IncreaseSQL):
                arguments.append(f"{key} = {key} + {self._format_parameter(key + '_datainput')}")
            elif isinstance(value, DecreaseSQL):
                arguments.append(f"{key} = {key} - {self._format_parameter(key + '_datainput')}")
            else:
                arguments.append(f"{key} = {self._format_parameter(key + '_datainput')}")
        return ", ".join(arguments)  # This removed the last comma in string

    def _split_where(self,
                     where: Optional[Union[str, List[str], Tuple[Union[List[str], str], Union[List, Dict]]]] = None) \
            -> Tuple[Union[str, List[str], None], Union[List, Dict, None]]:
        if not where:
            return (None, None)

        if not isinstance(where, tuple):
            return (where, None)

        if len(where) == 1:
            return (where[1], None)

        return (where[0], where[1])

    def _where(self, conditions: Optional[Union[List, str]] = None) -> str:
        if not conditions:
            return ""

        if isinstance(conditions, list):
            return f" WHERE {' AND '.join(conditions)}"

        return f" WHERE {conditions}"

    def _having(self, having: Optional[str] = None) -> str:
        if not having:
            return ""

        return f" HAVING {having}"

    def _with_sq(self, with_sq: Optional[Dict[str, str]] = None) -> str:
        if not with_sq:
            return ""

        return "WITH " + ", ".join(
            [f"{key} as ({value})" for key, value in with_sq.items()]
        )

    def _group(self, group: Optional[Union[List[str], str]] = None) -> str:
        if not group:
            return ""

        if isinstance(group, list):
            return f" GROUP BY {', '.join(group)}"

        return f" GROUP BY {group} "

    def _order(self, order: Optional[Union[str, Tuple[str, Union[Order, str]]]] = None) -> str:
        if not order:
            return ""

        if isinstance(order, str):
            return f" ORDER BY {order}"

        # If order is not a string, then it must be a tuple, here we just need to check the type of the 2º parameter
        if isinstance(order[1], str):
            return f" ORDER BY {order[0]} {order[1]}"

        # Second parameter must be of type Order
        return f" ORDER BY {order[0]} {order[1].value}"

    def _limit(self, limit: Optional[int]) -> str:
        if limit:
            return f" LIMIT {limit}"
        return ""

    def _offset(self, offset: Optional[int]) -> str:
        if offset:
            return f" OFFSET {offset}"
        return ""

    def _returning(self, returning: Optional[Union[str, List[str]]]) -> str:
        if not returning:
            return ""

        if isinstance(returning, list):
            return f" RETURNING {', '.join(returning)}"

        return f" RETURNING {returning}"

    def _fields(self, fields: Union[str, List[str]]) -> str:
        if isinstance(fields, str):
            return fields

        # It must be a list of strings
        return ', '.join(fields)

    def _select(
            self, table=None, fields=(), where=None, group=None, having=None, order=None, limit=None, offset=None,
            with_sq=None
    ) -> str:
        return (
                self._with_sq(with_sq)
                + f"SELECT {self._fields(fields)} FROM {table}"
                + self._where(where)
                + self._group(group)
                + self._having(having)
                + self._order(order)
                + self._limit(limit)
                + self._offset(offset)
        )

    def __del__(self):
        try:
            self._cursor.close()
        except Exception as e:  # TODO: Narrow down this exception
            if "Cannot operate on a closed database." not in str(e):
                raise e


class Psycopg2Sqlify(BaseSqlify):
    _unnamed_parameter = "%s"

    def _format_parameter(self, parameter: str) -> str:
        return f"%({parameter})s"


class Sqlite3Sqlify(BaseSqlify):
    _unnamed_parameter = "?"

    def _format_parameter(self, parameter: str) -> str:
        return f":{parameter}"



import sqlite3

from sqlify import Session

if __name__ == "__main__":
    conn = sqlite3.connect('test.db')

    with Session(conn, autocommit=True) as sqlify:
        rest = sqlify.fetchone(
            table="test",
            fields="column_1",
        )
        print(rest)
        #
        # rest = sqlify.insert(
        #     table="test",
        #     data={"column_1": "test2"}
        # )
        # sqlify.commit()


        # rest = sqlify.update(
        #     table="test",
        #     data={"column_1": "test3"},
        #     where=("column_1 = 'test2'", None)
        # )
        # sqlify.commit()
        print(rest)




from enum import Enum


class Order(Enum):
    ASC = "ASC"
    DESC = "DESC"


class DatabaseType(Enum):
    PSYCOPG2 = "PSYCOPG2"
    SQLITE3 = "SQLITE3"


class Fetch(Enum):
    ONE = "ONE"
    ALL = "ALL"



# -*- coding: utf-8 -*-
from typing import Optional, Union, Any, Type

from .builder import BaseSqlify, Psycopg2Sqlify, Sqlite3Sqlify
from .value_objects import DatabaseType

try:
    from psycopg2._psycopg import connection as psycopg2_connection
except ModuleNotFoundError:
    psycopg2_connection = None

try:
    from sqlite3 import Connection as sqlite3_connection, ProgrammingError
except ModuleNotFoundError:
    sqlite3_connection = None


class Session(object):
    _database_type: DatabaseType
    _manager: Type[BaseSqlify]

    def __init__(self, connection: Union[psycopg2_connection, sqlite3_connection],
                 database_type: Optional[DatabaseType] = None, autocommit: Optional[bool] = True):
        self._connection = connection
        self._autocommit = autocommit

        if database_type is not None:
            self._database_type = database_type
        elif psycopg2_connection and isinstance(self._connection, psycopg2_connection):
            self._database_type = DatabaseType.PSYCOPG2
            self._manager = Psycopg2Sqlify
        elif sqlite3_connection and isinstance(self._connection, sqlite3_connection):
            self._database_type = DatabaseType.SQLITE3
            self._manager = Sqlite3Sqlify
        else:
            raise RuntimeError(
                "Could not detect the correct database type, please supply the 'database_type' parameter")

        self.session = self._manager(self.get_cursor())

    @property
    def is_open(self) -> bool:
        if self._database_type == DatabaseType.PSYCOPG2:
            return not self._connection.closed
        elif self._database_type == DatabaseType.SQLITE3:
            return True
        raise NotImplementedError("Database type not implemented")

    def close(self) -> None:
        self._connection.close()

    def get_cursor(self) -> Any:
        return self._connection.cursor()

    def __enter__(self):
        return self.session

    def __exit__(self, type_, value, traceback):
        if self._autocommit:
            self._connection.commit()

        if self.is_open:
            self.close()
